use std::fmt::{Debug, Display};

use anyhow::{anyhow, Result};
use bytes::Bytes;
use futures::Stream;
use tokio_stream::StreamExt;

#[derive(Clone)]
pub struct Client {
    id: String,
    nats: async_nats::Client,
}

#[derive(Debug)]
pub enum Subject {
    Health,
    Exec,
    OpenTerminal,
}

impl Display for Subject {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Subject::Health => write!(f, "health"),
            Subject::Exec => write!(f, "exec"),
            Subject::OpenTerminal => write!(f, "open_terminal"),
        }
    }
}

impl TryFrom<&str> for Subject {
    type Error = anyhow::Error;

    fn try_from(value: &str) -> Result<Self> {
        Ok(match value {
            "health" => Subject::Health,
            "exec" => Subject::Exec,
            "open_terminal" => Subject::OpenTerminal,
            _ => return Err(anyhow!("unknown subject '{}'", value)),
        })
    }
}

#[derive(Debug)]
pub struct Subscriber {
    subject: Subject,
}

#[derive(Debug)]
pub struct Publisher {
    subject: Subject,
}

#[derive(Debug)]
pub struct Message {
    subject: Subject,
    payload: Bytes,
    subscribers: Vec<Subscriber>,
    publishers: Vec<Publisher>,
    reply: Option<async_nats::Subject>,
}

impl Message {
    fn from_transport(msg: async_nats::Message) -> Result<Self> {
        let suffix = msg.subject.split_terminator('.').last().unwrap_or_default();
        let subject = suffix.try_into()?;

        Ok(Message {
            subject,
            subscribers: Vec::default(),
            publishers: Vec::default(),
            payload: msg.payload,
            reply: msg.reply,
        })
    }

    pub fn subject(&self) -> &Subject {
        &self.subject
    }

    pub fn body(&self) -> Bytes {
        self.payload.clone()
    }

    pub fn health(health: crate::health::Health) -> Result<Message> {
        Ok(Message {
            subject: Subject::Health,
            payload: Bytes::from_iter(serde_json::to_vec(&health)?),
            reply: None,
        })
    }
}

impl Client {
    pub async fn connect(id: &str) -> Result<Self> {
        let nats = async_nats::ConnectOptions::new()
            .name(format!("Prymn Agent {id}"))
            .custom_inbox_prefix(format!("_INBOX_{id}"))
            .user_and_password("demo_agent".to_owned(), "demo_agent_password".to_owned())
            .connect("localhost")
            .await?;

        tracing::debug!("connected to NATS");

        Ok(Self {
            id: id.to_owned(),
            nats,
        })
    }

    pub async fn publish(&self, msg: Message) {
        let subject = format!("agents.v1.{}.{}", self.id, msg.subject);
        self.nats.publish(subject, msg.payload).await.unwrap();
    }

    pub async fn reply(&self, msg: Message, mut stream: impl Stream<Item = Bytes> + Unpin) {
        match msg.reply {
            Some(reply) => {
                while let Some(data) = stream.next().await {
                    self.nats.publish(reply.clone(), data.into()).await.unwrap();
                }
            }
            None => tracing::warn!(?msg, "tried to reply to message without a reply subject"),
        }
    }

    pub async fn subscribe(&self) -> Result<impl Stream<Item = Message>> {
        let subject = format!("agents.v1.{}.*", self.id);

        let subscriber = self
            .nats
            .subscribe(subject.clone())
            .await?
            .filter_map(move |data| match Message::from_transport(data) {
                Ok(msg) => Some(msg),
                Err(err) => {
                    tracing::warn!("{}", err);
                    None
                }
            });

        // let channel = sync::broadcast::channel(10);
        // while let Some(message) = subscriber.next().await {
        //     let subscriptions = self.further_subscriptions(message.subject).await;
        // }

        tracing::debug!("subscribed on {}", &subject);

        Ok(subscriber)
    }
}

impl Debug for Client {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Client {{ id: {} }}", self.id)
    }
}
